"""
Created on 19 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

https://opensensorsio.helpscoutdocs.com/article/84-overriding-timestamp-information-in-message-payload
"""

import optparse

from scs_core.osio.config.project import Project


# --------------------------------------------------------------------------------------------------------------------

class CmdOSIOTopicPublisher(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -t TOPIC | -c { C | G | P | S | X } } [-o] [-v]",
                                              version="%prog 1.0")

        # compulsory...
        self.__parser.add_option("--topic", "-t", type="string", nargs=1, action="store", dest="topic",
                                 help="topic path")

        self.__parser.add_option("--channel", "-c", type="string", nargs=1, action="store", dest="channel",
                                 help="publication channel")

        # optional...
        self.__parser.add_option("--override", "-o", action="store_true", dest="override", default=False,
                                 help="override OSIO reception datetime")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if bool(self.topic) == bool(self.channel):
            return False

        if self.channel and not Project.is_valid_channel(self.channel):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__opts.topic


    @property
    def channel(self):
        return self.__opts.channel


    @property
    def override(self):
        return self.__opts.override


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdOSIOTopicPublisher:{topic:%s, channel:%s, override:%s, verbose:%s}" % \
                    (self.topic, self.channel, self.override, self.verbose)



"""
Created on 26 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleSubset(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -i | -n | -s } { [-e EQUAL] | [-l LOWER] [-u UPPER] } "
                                                    "[-t] [-x] [-v] PATH", version="%prog 1.0")

        # casting...
        self.__parser.add_option("--iso8601", "-i", action="store_true", dest="iso8601", default=False,
                                 help="interpret the value as an ISO 8601 datetime")

        self.__parser.add_option("--numeric", "-n", action="store_true", dest="numeric", default=False,
                                 help="interpret the value as a number")

        self.__parser.add_option("--string", "-s", action="store_true", dest="string", default=False,
                                 help="interpret the value as a string")

        self.__parser.add_option("--strict", "-t", action="store_true", dest="strict", default=False,
                                 help="halt on type errors")

        # function...
        self.__parser.add_option("--equal", "-e", type="string", nargs=1, action="store", dest="equal",
                                 help="equal to")

        self.__parser.add_option("--lower", "-l", type="string", nargs=1, action="store", dest="lower",
                                 help="lower bound")

        self.__parser.add_option("--upper", "-u", type="string", nargs=1, action="store", dest="upper",
                                 help="upper bound")

        # output...
        self.__parser.add_option("--exclusions", "-x", action="store_true", dest="exclusions", default=False,
                                 help="output exclusions instead of inclusions")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.iso8601:
            count += 1

        if self.numeric:
            count += 1

        if self.string:
            count += 1

        if count != 1:
            return False

        if self.equal is not None and (self.lower is not None or self.upper is not None):
            return False

        if self.lower is not None and self.upper is not None and self.upper <= self.lower:
            return False

        if len(self.__args) != 1:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def inclusions(self):
        return not self.exclusions


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def iso8601(self):
        return self.__opts.iso8601


    @property
    def numeric(self):
        return self.__opts.numeric


    @property
    def string(self):
        return self.__opts.string


    @property
    def strict(self):
        return self.__opts.strict


    @property
    def equal(self):
        return self.__cast(self.__opts.equal)


    @property
    def lower(self):
        return self.__cast(self.__opts.lower)


    @property
    def upper(self):
        return self.__cast(self.__opts.upper)


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def exclusions(self):
        return self.__opts.exclusions


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def __cast(self, value):
        if value is None:
            return None

        if self.string:
            return str(value)

        cast_value = LocalizedDatetime.construct_from_iso8601(value) if self.iso8601 else Datum.float(value)

        if cast_value is None:
            raise ValueError(value)

        return cast_value


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleSubset:{iso8601:%s, numeric:%s, string:%s, equal:%s, lower:%s, upper:%s, " \
               "strict:%s, exclusions:%s, verbose:%s, path:%s}" % \
               (self.iso8601, self.numeric, self.string, self.__opts.equal, self.__opts.lower, self.__opts.upper,
                self.strict, self.exclusions, self.verbose, self.path)



"""
Created on 9 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

uses matplotlibrc configuration file

https://matplotlib.org/faq/usage_faq.html
https://stackoverflow.com/questions/28269157/plotting-in-a-non-blocking-way-with-matplotlib?noredirect=1&lq=1
http://bastibe.de/2013-05-30-speeding-up-matplotlib.html
"""

from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator

from scs_analysis.chart.chart import Chart

from scs_core.data.histogram import Histogram


# --------------------------------------------------------------------------------------------------------------------

class HistoChart(Chart):
    """
    classdocs
    """

    def __init__(self, title, batch_mode, x_min, x_max, bin_count, precision, path, outfile=None):
        """
        Constructor
        """
        Chart.__init__(self, title, batch_mode)

        # fields...
        self.__x_min = x_min
        self.__x_max = x_max
        self.__y_max = 1

        self.__bin_count = bin_count

        self.__path = path

        self.__outfile = outfile

        # histo...
        self.__histogram = Histogram(x_min, x_max, bin_count, precision, path)

        # plotter...
        self.__fig, self.__ax = plt.subplots()
        self.__fig.canvas.set_window_title(self.title(path))
        self.__fig.tight_layout()

        self.__fig.canvas.mpl_connect('close_event', self.close)

        self.__ax.set(xlim=[self.__x_min, self.__x_max], ylim=[0, self.__y_max])
        self.__ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        self.__ax.yaxis.grid(True)

        if not batch_mode:
            domain_width = abs(self.__x_max - self.__x_min)
            bar_width = domain_width / bin_count

            self.__rects = plt.bar(self.__histogram.bins, self.__histogram.counts, bar_width)


    # ----------------------------------------------------------------------------------------------------------------

    def plot(self, path_dict):
        if self.closed:
            return

        # datum...
        try:
            value = path_dict.node(self.__path)
        except KeyError:
            return

        if value is None:
            return

        try:
            datum = float(value)

            # compute...
            index, count = self.__histogram.append(datum)
            max_count = self.__histogram.max_count

            if not self.batch_mode:
                self.__rects[index].set_height(count)

            # render...
            if max_count > self.__y_max:
                self.__y_max = max_count
                self.__ax.set(ylim=[0, self.__y_max + (self.__y_max * 0.10)])

                if not self.batch_mode:
                    plt.pause(0.001)
            else:
                if not self.batch_mode:
                    self.__ax.draw_artist(self.__rects[index])
                    self.__fig.canvas.flush_events()                # may need self.__fig.canvas.update()

        except ValueError:
            pass


    def render(self, block=True):
        if self.batch_mode:
            domain_width = abs(self.__x_max - self.__x_min)
            bar_width = domain_width / self.__bin_count

            plt.bar(self.__histogram.bins, self.__histogram.counts, bar_width)

        plt.draw()
        plt.show(block=block)
        plt.pause(0.001)


    def close(self, _event):
        self._closed = True

        if self.__outfile is None:
            return

        self.__histogram.to_csv(self.__outfile)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def x_min(self):
        return self.__x_min


    @property
    def x_max(self):
        return self.__x_max


    @property
    def y_max(self):
        return self.__y_max


    @property
    def outfile(self):
        return self.__outfile


    @property
    def histogram(self):
        return self.__histogram


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "HistoChart:{batch_mode:%s, x_min:%0.3f, x_max:%0.3f, y_max:%0.3f, path:%s, outfile:%s}" % \
                (self.batch_mode, self.x_min, self.x_max, self.y_max, self.__path, self.outfile)



"""
Created on 17 Jan 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import logging
import time

from scs_core.data.timedelta import Timedelta
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class BatchDownloadReporter(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        """
        Constructor
        """
        self.__block_count = 0
        self.__document_count = 0
        self.__start_time = time.time()

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def reset(self):
        self.__block_count = 0
        self.__document_count = 0
        self.__start_time = time.time()


    def print(self, block_length, block_start=None):
        if Logging.level() > logging.INFO:
            return

        self.__block_count += 1
        self.__document_count += block_length
        elapsed_time = int(round(time.time() - self.__start_time))
        elapsed_delta = Timedelta(seconds=elapsed_time)
        elapsed = elapsed_delta.as_json()

        if block_start is None:
            self.__logger.info("docs:%d elapsed:%s" % (self.__document_count, elapsed))
        else:
            self.__logger.info("block start:%s docs:%d elapsed:%s" % (block_start, self.__document_count, elapsed))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def block_count(self):
        return self.__block_count


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "BatchDownloadReporter:{block_count:%d, document_count:%d, start_time:%d}" % \
               (self.block_count, self.__document_count, self.__start_time)



#!/usr/bin/env python3

"""
Created on 4 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The aws_mqtt_client utility is used to subscribe or publish using the South Coast Science / AWS messaging
infrastructure.

Documents for publication are gained from stdin by default, otherwise from the specified Unix domain socket (UDS).
Likewise, documents gained from subscription are written to stdout, or a specified UDS.

Subscriptions can be specified either by a project channel name, or by an explicit messaging topic path. Documents
gained by subscription may be delivered either to stdout, or to a specified Unix domain socket.

The --timed mode is provided so that the difference between the payload's recorded ('rec') datetime and the
received datetime can be monitored. This is useful in assessing network performance.

The aws_mqtt_client utility requires the AWS client authorisation to operate.

Only one MQTT client should run at any one time, per TCP/IP host.

SYNOPSIS
Usage: aws_mqtt_client.py [-p UDS_PUB] [-s] { -c { C | G | P | S | X } (UDS_SUB_1) |
[SUB_TOPIC_1 (UDS_SUB_1) .. SUB_TOPIC_N (UDS_SUB_N)] } [-n] [-t] [-e] [-v]

EXAMPLES
aws_mqtt_client.py -n south-coast-science-dev/production-test/loc/1/gases
aws_mqtt_client.py -nt south-coast-science-dev/cube/loc/1/climate | node.py received payload.rec

FILES
~/SCS/aws/aws_client_auth.json
~/SCS/aws/certs/NNNNNNNNNN-certificate.pem.crt
~/SCS/aws/certs/NNNNNNNNNN-private.pem.key
~/SCS/aws/certs/NNNNNNNNNN-public.pem.key
~/SCS/aws/certs/root-CA.crt

SEE ALSO
scs_analysis/aws_client_auth

BUGS
When run as a background process, aws_mqtt_client will exit if it has no stdin stream.
"""

import json
import sys

from scs_analysis.cmd.cmd_mqtt_client import CmdMQTTClient

from scs_analysis.handler.aws_mqtt_publisher import AWSMQTTPublisher
from scs_analysis.handler.aws_mqtt_subscription_handler import AWSMQTTSubscriptionHandler
from scs_analysis.handler.mqtt_reporter import MQTTReporter

from scs_core.aws.client.client_auth import ClientAuth
from scs_core.aws.client.mqtt_client import MQTTClient, MQTTSubscriber
from scs_core.aws.config.project import Project

from scs_core.comms.mqtt_conf import MQTTConf
from scs_core.comms.uds_reader import UDSReader
from scs_core.comms.uds_writer import UDSWriter

from scs_core.data.publication import Publication

from scs_core.sys.filesystem import Filesystem
from scs_core.sys.system_id import SystemID

from scs_host.comms.domain_socket import DomainSocket
from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    conf = None
    source = None
    reporter = None
    publisher = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdMQTTClient()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("aws_mqtt_client: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # MQTTConf
        conf = MQTTConf.load(Host, skeleton=True)

        if cmd.verbose:
            print("aws_mqtt_client: %s" % conf, file=sys.stderr)

        # ClientAuth...
        auth = ClientAuth.load(Host)

        if auth is None:
            print("aws_mqtt_client: ClientAuth not available.", file=sys.stderr)
            exit(1)

        # comms...
        source = UDSReader(DomainSocket, cmd.uds_pub)

        if cmd.verbose:
            print("aws_mqtt_client: %s" % source, file=sys.stderr)

        # reporter...
        reporter = MQTTReporter(cmd.verbose)

        # subscribers...
        subscribers = []

        if cmd.channel:
            # SystemID...
            system_id = SystemID.load(Host)

            if system_id is None:
                print("aws_mqtt_client: SystemID not available.", file=sys.stderr)
                exit(1)

            if cmd.verbose:
                print("aws_mqtt_client: %s" % system_id, file=sys.stderr)

            # Project...
            project = Project.load(Host)

            if project is None:
                print("aws_mqtt_client: Project not available.", file=sys.stderr)
                exit(1)

            if cmd.verbose:
                print("aws_mqtt_client: %s" % project, file=sys.stderr)

            topic = project.channel_path(cmd.channel, system_id)

            # subscriber...
            sub_comms = UDSWriter(DomainSocket, cmd.channel_uds)

            handler = AWSMQTTSubscriptionHandler(reporter, comms=sub_comms,
                                                 wrap=cmd.wrap, timed=cmd.timed, echo=cmd.echo)

            subscribers.append(MQTTSubscriber(topic, handler.handle))

        else:
            for subscription in cmd.subscriptions:
                sub_comms = UDSWriter(DomainSocket, subscription.address)

                # subscriber...
                handler = AWSMQTTSubscriptionHandler(reporter, comms=sub_comms,
                                                     wrap=cmd.wrap, timed=cmd.timed, echo=cmd.echo)

                if cmd.verbose:
                    print("aws_mqtt_client: %s" % handler, file=sys.stderr)

                subscribers.append(MQTTSubscriber(subscription.topic, handler.handle))

        # client...
        client = MQTTClient(*subscribers)
        publisher = AWSMQTTPublisher(conf, auth, client, reporter)

        if cmd.verbose:
            print("aws_mqtt_client: %s" % publisher, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # client...
        if not conf.inhibit_publishing:
            publisher.connect()

        # data source...
        source.connect()

        # process input...
        for message in source.messages():
            # receive...
            try:
                jdict = json.loads(message)

            except (TypeError, ValueError) as ex:
                reporter.print("%s: %s" % (ex, message))
                continue

            if cmd.echo:
                print(message)
                sys.stdout.flush()

            if conf.inhibit_publishing:
                continue

            publication = Publication.construct_from_jdict(jdict)

            if publication is None:
                continue

            # publish...
            publisher.publish(publication)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except ConnectionError as ex:
        print("aws_mqtt_client: %s" % ex, file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd and cmd.verbose:
            print("aws_mqtt_client: finishing", file=sys.stderr)

        if source:
            source.close()

        if publisher:
            publisher.disconnect()

        if conf:
            Filesystem.rm(conf.report_file)



"""
Created on 19 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleTimezone(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -z | [-i ISO_PATH] TIMEZONE_NAME }", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--zones", "-z", action="store_true", dest="zones", default=False,
                                 help="list the available timezone names to stderr")

        self.__parser.add_option("--iso-path", "-i", type="string", nargs=1, action="store", default="rec", dest="iso",
                                 help="path for ISO 8601 datetime output (default 'rec')")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if bool(self.timezone) == bool(self.__opts.zones):
            return False

        if self.iso is not None and self.timezone is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def zones(self):
        return self.__opts.zones


    @property
    def iso(self):
        return self.__opts.iso


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def timezone(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleTimezone:{zones:%s, iso:%s, verbose:%s, timezone:%s}" % \
               (self.zones, self.iso, self.verbose, self.timezone)



"""
Created on 13 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdMultiChart(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-b] [-x POINTS] [-y MIN MAX] [-e] [-t TITLE] [-v] "
                                                    "PATH_1 .. PATH_N", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--batch", "-b", action="store_true", dest="batch_mode", default=False,
                                 help="wait for all data before displaying chart")

        self.__parser.add_option("--x", "-x", type="int", nargs=1, action="store", default=600, dest="x",
                                 help="number of x points (default 600)")

        self.__parser.add_option("--y", "-y", type="float", nargs=2, action="store", default=(-10.0, 10.0), dest="y",
                                 help="set y-axis to min / max (default -10, 10)")

        self.__parser.add_option("--echo", "-e", action="store_true", dest="echo", default=False,
                                 help="echo stdin to stdout")

        self.__parser.add_option("--skip-malformed", "-s", action="store_true", dest="skip_malformed", default=False,
                                 help="ignore rows with missing path values")

        self.__parser.add_option("--title", "-t", type="string", nargs=1, action="store", dest="title",
                                 help="chart title")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if len(self.paths) == 0:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def batch_mode(self):
        return self.__opts.batch_mode


    @property
    def x(self):
        return self.__opts.x


    @property
    def y(self):
        return self.__opts.y


    @property
    def echo(self):
        return self.__opts.echo


    @property
    def skip_malformed(self):
        return self.__opts.skip_malformed


    @property
    def title(self):
        return self.__opts.title


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def paths(self):
        return self.__args


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdMultiChart:{batch_mode:%s, x:%d, y:%s, echo:%s, skip_malformed:%s, title:%s, " \
               "verbose:%s, paths:%s}" % \
                    (self.batch_mode, self.x, self.y, self.echo, self.skip_malformed, self.title,
                     self.verbose, self.paths)



"""
Created on 29 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.aws.manager.alert_status_finder import AlertStatusFinderRequest


# --------------------------------------------------------------------------------------------------------------------

class CmdAlertStatus(object):
    """unix command line handler"""

    __CAUSES = {'L': '<L', 'U': '>U', 'N': 'NV'}

    # --------------------------------------------------------------------------------------------------------------------

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -l | -d [-c CAUSE] } [-i INDENT] [-v] ID",
                                              version="%prog 1.0")

        # operations...
        self.__parser.add_option("--latest", "-l", action="store_true", dest="latest", default=False,
                                 help="return latest status report only")

        self.__parser.add_option("--history", "-d", action="store_true", dest="history", default=False,
                                 help="return history of status reports")

        # filters...
        self.__parser.add_option("--cause", "-c", type="string", action="store", dest="cause",
                                 help="filter by cause { L | U | N }")

        # output...
        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.id is None:
            return False

        if self.latest == self.history:
            return False

        if self.latest and self.__opts.cause is not None:
            return False

        if self.__opts.cause is not None and self.__opts.cause not in self.__CAUSES:
            return False

        return True


    def response_mode(self):
        return AlertStatusFinderRequest.Mode.LATEST if self.latest else AlertStatusFinderRequest.Mode.HISTORY


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__args[0] if self.__args else None


    @property
    def latest(self):
        return self.__opts.latest


    @property
    def history(self):
        return self.__opts.history


    @property
    def cause(self):
        return None if self.__opts.cause is None else self.__CAUSES[self.__opts.cause]


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdAlertStatus:{id:%s, latest:%s, history:%s, cause:%s, indent:%s, verbose:%s}" % \
               (self.id, self.latest, self.history, self.__opts.cause, self.indent, self.verbose)



#!/usr/bin/env python3

"""
Created on 15 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_concentration utility is used to inject a gas concentration value, based on a given density (microgrammes
per cubic metre), temperature and pressure. The reported value is in parts per billion.

Density and temperature must be supplied from the input JSON document. Atmospheric pressure may be supplied from the
input JSON document. Alternatively, a pressure estimate may be supplied on the command line. In both cases, pressure
should be indicated in kilopascals (kPa). If no pressure value is supplied, the standard pressure of 101.3 kPa is used.

The path for the node in the output document indicating concentration is based on the path for the node in the
input document indicating density. For example, if the input path is SUB-PATH.dns, the output path is SUB-PATH.cnc. If
a SUB-PATH.cnc field exists in the input document, it is overwritten.

SYNOPSIS
sample_concentration.py [-v] GAS DENSITY_PATH T_PATH [{P_PATH | -p PRESSURE}]

EXAMPLES
csv_reader.py joined_2019-02.csv | sample_concentration.py -v NO2 ref.val.NO2.dns praxis.val.sht.tmp

DOCUMENT EXAMPLE - INPUT
{"rec": "2019-02-12T13:00:00Z", "praxis": {"val":
{"NO2": {"weV": 0.300059, "cnc": 23.4, "aeV": 0.302339, "weC": -0.002793},
"sht": {"hmd": 67.6, "tmp": 13.1}}}, "ref": {"val": {"NO2": {"units": "ugm-3", "dns": 51}}}}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2019-02-12T13:00:00Z", "praxis": {"val":
{"NO2": {"weV": 0.300059, "cnc": 23.4, "aeV": 0.302339, "weC": -0.002793},
"sht": {"hmd": 67.6, "tmp": 13.1}}}, "ref": {"val": {"NO2": {"units": "ugm-3", "dns": 51, "cnc": 26.0}}}}

RESOURCES
http://www.apis.ac.uk/unit-conversion
"""

import sys

from scs_analysis.cmd.cmd_sample_concentration import CmdSampleConcentration

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict

from scs_core.gas.gas import Gas


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    density = None
    t = None
    p = None

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleConcentration()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if not Gas.is_valid_name(cmd.gas):
        print("sample_concentration: the gas '%s' is not recognised." % cmd.gas, file=sys.stderr)
        exit(1)

    if cmd.verbose:
        print("sample_concentration: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        density_nodes = cmd.density_path.split('.')
        concentration_path = '.'.join(density_nodes[:-1] + ['cnc'])


        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            jstr = line.strip()
            datum = PathDict.construct_from_jstr(jstr)

            if datum is None:
                continue

            document_count += 1

            paths = datum.paths()

            # density / t...
            if cmd.density_path not in paths or cmd.t_path not in paths:
                continue

            density_node = datum.node(cmd.density_path)
            t_node = datum.node(cmd.t_path)

            if density_node == '' or t_node == '':
                continue

            try:
                density = float(density_node)
            except ValueError:
                print("sample_concentration: invalid value for density in %s" % jstr, file=sys.stderr)
                exit(1)

            try:
                t = float(t_node)
            except ValueError:
                print("sample_concentration: invalid value for t in %s" % jstr, file=sys.stderr)
                exit(1)

            # p...
            if cmd.p_path:
                p_node = datum.node(cmd.p_path)

                if p_node == '':
                    continue

                try:
                    p = float(p_node)
                except ValueError:
                    print("sample_concentration: invalid value for p in %s" % jstr, file=sys.stderr)
                    exit(1)

            else:
                p = cmd.pressure

            # compute...
            cnc = round(Gas.concentration(cmd.gas, density, t, p), 1)

            # copy...
            target = PathDict()

            for path in paths:
                if path == concentration_path:
                    continue

                target.append(path, datum.node(path))

                if path == cmd.density_path:
                    target.append(concentration_path, cnc)

            # report...
            print(JSONify.dumps(target))
            sys.stdout.flush()

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_concentration: KeyError: %s" % ex, file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_concentration: documents: %d processed: %d" % (document_count, processed_count),
                  file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 6 Nov 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The aws_topic_history utility is used to retrieve stored data from the South Coast Science / AWS historic data
retrieval system. Data can be retrieved by start or start + end localised datetimes, or by a days / hours / minutes
timedelta back in time from now.

Where start and end datetimes are used, messages are retrieved such that:

start datetime <= message rec datetime < end datetime

A latest mode returns the most recent document, or none if the topic has never received a publication. A --latest-at
mode returns the most recent document, or none, prior to, or at the given datetime.

The --rec-only flag causes only the rec fields on the documents to be returned. This results in much faster data
retrieval, and is useful if sampling continuity is being tested.

Note that no check is made for the existence of the topic - if the topic does not exist, then no error is raised and
no data is returned.

Equivalent to cURL:
curl "https://aws.southcoastscience.com/topicMessages?topic=south-coast-science-dev/production-test/loc/1/gases
&startTime=2018-12-13T07:03:59.712Z&endTime=2018-12-13T15:10:59.712Z"

SYNOPSIS
aws_topic_history.py { -l | -a LATEST_AT | -t { [[DD-]HH:]MM[:SS] | :SS } | -s START [-e END] }
{ -c HH:MM:SS [-m] [-x] | [-w] [-f] } [-r] [{ -v | -d }] TOPIC

EXAMPLES
aws_topic_history.py south-coast-science-dev/production-test/loc/1/gases -t 1 -v -w

DOCUMENT EXAMPLE - OUTPUT
{"device": "scs-bbe-401", "topic": "south-coast-science-demo/brighton/loc/1/climate", "upload": "2019-01-11T12:15:36Z",
"payload": {"val": {"hmd": 68.4, "tmp": 12.3}, "rec": "2019-01-11T12:15:36Z", "tag": "scs-bgx-401"}}

{"val": {"hmd": 68.4, "tmp": 12.3}, "rec": "2019-01-11T12:15:36Z", "tag": "scs-bgx-401"}

FILES
~/SCS/aws/aws_api_auth.json

SEE ALSO
scs_analysis/aws_api_auth
scs_analysis/localised_datetime

RESOURCES
https://github.com/curl/curl
"""

import requests
import sys

from scs_analysis.cmd.cmd_aws_topic_history import CmdAWSTopicHistory
from scs_analysis.handler.batch_download_reporter import BatchDownloadReporter

from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.manager.byline_manager import BylineManager
from scs_core.aws.manager.lambda_message_manager import MessageManager

from scs_core.client.network import Network
from scs_core.client.resource_unavailable_exception import ResourceUnavailableException

from scs_core.data.checkpoint_generator import CheckpointGenerator
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    agent = None
    reporter = None
    start_time = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdAWSTopicHistory()

    Logging.config('aws_topic_history', level=cmd.log_level())
    logger = Logging.getLogger()

    if not cmd.is_valid_latest_at():
        logger.error("invalid format for latest-to datetime.")
        exit(2)

    if not cmd.is_valid_timedelta():
        logger.error("invalid format for timedelta.")
        exit(2)

    if not cmd.is_valid_start():
        logger.error("invalid format for start datetime.")
        exit(2)

    if not cmd.is_valid_end():
        logger.error("invalid format for end datetime.")
        exit(2)

    if cmd.checkpoint is not None and cmd.checkpoint != 'auto' and not CheckpointGenerator.is_valid(cmd.checkpoint):
        logger.error("invalid format for checkpoint.")
        exit(2)

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # APIAuth...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            logger.error("APIAuth not available.")
            exit(1)

        logger.info(api_auth)

        # reporter...
        reporter = BatchDownloadReporter()

        # BylineManager...
        byline_manager = BylineManager(requests)

        # MessageManager...
        message_manager = MessageManager(api_auth, reporter=reporter)

        logger.info(message_manager)

        # ------------------------------------------------------------------------------------------------------------
        # check...

        if not Network.is_available():
            logger.info("waiting for network.")
            Network.wait()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        start_time = LocalizedDatetime.now()

        if cmd.latest_at:
            message = message_manager.find_latest_for_topic(cmd.topic, cmd.latest_at, None, cmd.include_wrapper,
                                                            cmd.rec_only, None)

            if message:
                print(JSONify.dumps(message))

            exit(0)

        # start / end times...
        if cmd.latest:
            byline = byline_manager.find_latest_byline_for_topic(cmd.topic)

            if byline is None:
                exit(0)

            end = byline.rec.timedelta(seconds=1)
            start = byline.rec.timedelta(seconds=-1)        # TODO: should not need this

        elif cmd.timedelta:
            end = LocalizedDatetime.now()
            start = end - cmd.timedelta

        else:
            end = LocalizedDatetime.now() if cmd.end is None else cmd.end
            start = cmd.start

        logger.info("start: %s" % start)
        logger.info("end: %s" % end)

        # messages...
        for message in message_manager.find_for_topic(cmd.topic, start, end, None, cmd.fetch_last, cmd.checkpoint,
                                                      cmd.include_wrapper, cmd.rec_only, cmd.min_max,
                                                      cmd.exclude_remainder, False, None):
            print(JSONify.dumps(message))
            sys.stdout.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except (ConnectionError, HTTPException) as ex:
        logger.error("%s: %s" % (ex.__class__.__name__, ex))
        exit(1)

    except ResourceUnavailableException as ex:
        logger.error(repr(ex))
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if reporter:
            logger.info("blocks: %s" % reporter.block_count)

        if start_time:
            elapsed_time = LocalizedDatetime.now() - start_time
            logger.info("elapsed time: %s" % elapsed_time.as_json())



#!/usr/bin/env python3

"""
Created on 4 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The csv_reader utility is used to convert data from comma-separated value (CSV) format to JSON format.

The names of columns given in the CSV header row indicate paths into the JSON document: dictionary fields are separated
from their container by a period ('.') character, and JSON array members separated from their container by a
colon (':') character.

The first row of the CSV file (or stdin input) is assumed to be a header row. If there are more columns in the body of
the CSV than in the header, excess values are ignored. Header cells may not be empty.

By default, output is in the form of a sequence of JSON documents, separated by newlines. If the array (-a) option is
selected, output is in the form of a JSON array - the output opens with a '[' character, documents are separated by
the ',' character, and the output is terminated by a ']' character.

SYNOPSIS
csv_reader.py [-s] [-n] [-l LIMIT] [-a] [-v] [FILENAME_1 .. FILENAME_N]

EXAMPLES
csv_reader.py -v scs-ph1-10-status-2019-07-*.csv

DOCUMENT EXAMPLE - INPUT
tag,rec,val.hmd,val.tmp
scs-ap1-6,2018-04-04T14:50:38.394+00:00,59.7,23.8
scs-ap1-6,2018-04-04T14:55:38.394+00:00,59.8,23.9

DOCUMENT EXAMPLE - OUTPUT
Sequence mode:
{"tag": "scs-ap1-6", "rec": "2018-04-04T14:50:38.394+00:00", "val": {"hmd": 59.7, "tmp": 23.8}}
{"tag": "scs-ap1-6", "rec": "2018-04-04T14:55:38.394+00:00", "val": {"hmd": 59.8, "tmp": 23.9}}

Array mode:
[{"tag": "scs-ap1-6", "rec": "2018-04-04T14:50:38.394+00:00", "val": {"hmd": 59.7, "tmp": 23.8}},
{"tag": "scs-ap1-6", "rec": "2018-04-04T14:55:38.394+00:00", "val": {"hmd": 59.8, "tmp": 23.9}}]

SEE ALSO
scs_analysis/csv_writer
"""

import sys

from scs_analysis.cmd.cmd_csv_reader import CmdCSVReader

from scs_core.csv.csv_reader import CSVReader, CSVReaderException
from scs_core.csv.csv_dict import CSVHeaderError


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    file_count = 0
    total_rows = 0

    reader = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdCSVReader()

    if cmd.verbose:
        print("csv_reader: %s" % cmd, file=sys.stderr)

    if cmd.array:
        print('[', end='')

    try:
        for filename in cmd.filenames:

            file_count += 1
            rows = 0

            # --------------------------------------------------------------------------------------------------------
            # resources...

            try:
                reader = CSVReader.construct_for_file(filename, cast=cmd.cast, nullify=cmd.nullify)

            except FileNotFoundError:
                print("csv_reader: file not found: %s" % filename, file=sys.stderr)
                exit(1)

            except KeyError as ex:
                print("csv_reader: empty header cell in: %s." % ex, file=sys.stderr)
                exit(1)

            except ValueError as ex:
                print("csv_reader: duplicate column names in: %s." % ex, file=sys.stderr)
                exit(1)

            if cmd.verbose:
                print("csv_reader: %s" % reader, file=sys.stderr)
                sys.stderr.flush()


            # --------------------------------------------------------------------------------------------------------
            # run...

            try:
                for datum in reader.rows():
                    if cmd.limit is not None and rows >= cmd.limit:
                        break

                    if cmd.array:
                        if rows == 0:
                            print(datum, end='')

                        else:
                            print(", %s" % datum, end='')

                    else:
                        print(datum)

                    sys.stdout.flush()

                    rows += 1

            except CSVHeaderError as ex:
                print("csv_reader: clashing column names: '%s' and '%s'" % (ex.left, ex.right), file=sys.stderr)
                exit(1)

            except CSVReaderException as ex:
                if cmd.verbose:
                    print("csv_reader: ending file on row %d: %s" % (rows, ex), file=sys.stderr)
                    continue

            finally:
                if reader is not None:
                    reader.close()

            if cmd.verbose:
                print("csv_reader: rows: %d" % rows, file=sys.stderr)

            total_rows += rows


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.array:
            print(']')

        if cmd and cmd.verbose and file_count > 1:
            print("csv_reader: files: %d total rows: %d" % (file_count, total_rows), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 3 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The histo_chart utility is used to create Matplotlib histogram charts and comma-separated value (CSV) histogram files.
The utility analyses a given path to a leaf node of the input JSON data stream.

For the domain MIN and MAX flag, a value is included if:

MIN <= value < MAX

An optional "batch" ("-b") flag can be set, causing the plotting only to take place when all data points have been
received.

Depending on operating system, it may be necessary to use a matplotlibrc file, which specifies the Matplotlib
back-end graphics system.

SYNOPSIS
histo_chart.py [-b] [-x MIN MAX] [-c BIN_COUNT] [-p PRECISION] [-o FILENAME] [-e] [-t TITLE] [-v] PATH

EXAMPLES
socket_receiver.py | histo_chart.py val.CO2.cnc -x -10 10 -e -o CO2.csv

FILES
~/SCS/scs_analysis/src/scs_analysis/matplotlibrc

SEE ALSO
scs_analysis/multi_chart
scs_analysis/single_chart

BUGS
On some operating systems, the chart will remain as the uppermost window until all data have been received.
"""

import sys
import warnings

from scs_analysis.chart.histo_chart import HistoChart
from scs_analysis.cmd.cmd_histo_chart import CmdHistoChart

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict

from scs_core.sync.line_reader import LineReader

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    warnings.filterwarnings("ignore", module="matplotlib")

    chart = None
    proc = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdHistoChart()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('histo_chart', verbose=cmd.verbose)
    logger = Logging.getLogger()

    if cmd.bin_count < 1:
        logger.error("bin count must be greater than 0.")
        exit(2)

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # reader...
        reader = LineReader(sys.stdin.fileno())
        logger.info(reader)

        # chart...
        chart = HistoChart(cmd.title, cmd.batch_mode, cmd.x[0], cmd.x[1], cmd.bin_count, cmd.precision, cmd.path,
                           outfile=cmd.outfile)
        logger.info(chart)
        logger.info("backend: %s" % chart.backend())

        logger.error("terminate this utility by closing the chart window.")


        # ------------------------------------------------------------------------------------------------------------
        # run...

        proc = reader.start()

        for line in reader.lines:
            if chart.closed:
                break

            if line is None:
                chart.pause()
                continue

            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            if cmd.path not in datum.paths():
                if not cmd.skip_malformed:
                    logger.error("path: %s not in %s" % (cmd.path, line.strip()))
                    exit(1)

                continue

            try:
                float(datum.node(cmd.path))
            except ValueError:
                logger.error("invalid value for path: %s: %s" % (cmd.path, line.strip()))
                exit(1)

            if cmd.echo:
                print(JSONify.dumps(datum))
                sys.stdout.flush()

            chart.plot(datum)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if not chart.closed:
            chart.render()

    except KeyboardInterrupt:
        print(file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # close...

    finally:
        if proc:
            proc.terminate()

        if chart is not None and not chart.closed:
            logger.info("closing.")
            chart.close(None)



"""
Created on 10 Mar 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

from collections import OrderedDict

from scs_core.csv.csv_writer import CSVWriter

from scs_core.data.json import JSONable
from scs_core.data.str import Str
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class CSVSegmentor(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, max_interval: Timedelta, file_prefix):
        """
        Constructor
        """
        self.__max_interval = max_interval                      # timedelta
        self.__file_prefix = file_prefix                        # string
        self.__blocks = []                                      # array of CSVContiguousBlock


    def __len__(self):
        return len(self.__blocks)


    # ----------------------------------------------------------------------------------------------------------------

    def collate(self, rec, jstr):
        # initialise...
        if not self.__blocks:
            self.__blocks.append(CSVContiguousBlock.construct(None, rec, self.file_prefix))

        current_block = self.__blocks[-1]

        interval = None if current_block.end is None else rec - current_block.end

        # end of block...
        if interval is not None and interval > self.max_interval:
            current_block.close()

            # new block...
            current_block = CSVContiguousBlock.construct(interval, rec, self.file_prefix)
            self.__blocks.append(current_block)

        # append...
        current_block.append(interval, rec, jstr)


    def close(self):
        for block in self.__blocks:
            block.close()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def max_interval(self):
        return self.__max_interval


    @property
    def file_prefix(self):
        return self.__file_prefix


    @property
    def blocks(self):
        return self.__blocks


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVSegmentor:{max_interval:%s, file_prefix:%s, blocks:%s}" % \
               (self.max_interval, self.file_prefix, Str.collection(self.blocks))


# --------------------------------------------------------------------------------------------------------------------

class CSVContiguousBlock(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # e.g. scs-bgx-431-ref-particulates-N2-climate-2019_15min_clipped_2020-03-10T16-00-18Z.csv

    @classmethod
    def construct(cls, prev_interval, start, file_prefix):
        if file_prefix is None:
            return cls(prev_interval, start, None)

        file_name = '-'.join((file_prefix,  start.as_iso8601().replace(':', '-'))) + '.csv'
        return cls(prev_interval, start, CSVWriter(file_name))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, prev_interval, start, writer):
        """
        Constructor
        """
        self.__start = start                            # LocalizedDatetime
        self.__writer = writer                          # CSVWriter or None

        self.__prev_interval = prev_interval            # timedelta
        self.__max_interval = None                      # timedelta

        self.__end = None                               # LocalizedDatetime
        self.__count = 0                                # int


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, interval, rec, jstr):
        if self.__writer:
            self.__writer.write(jstr)

        if self.count > 0 and (self.__max_interval is None or interval > self.__max_interval):
            self.__max_interval = interval

        self.__end = rec
        self.__count += 1


    def close(self):
        if self.__writer:
            self.__writer.close()


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['start'] = self.start.as_json()
        jdict['end'] = self.end.as_json()

        jdict['prev-interval'] = None if self.prev_interval is None else \
            Timedelta.construct(self.prev_interval).as_json()

        jdict['max-interval'] = None if self.max_interval is None else \
            Timedelta.construct(self.max_interval).as_json()

        jdict['count'] = self.count

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def start(self):
        return self.__start


    @property
    def end(self):
        return self.__end


    @property
    def prev_interval(self):
        return self.__prev_interval


    @property
    def max_interval(self):
        return self.__max_interval


    @property
    def count(self):
        return self.__count


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVContiguousBlock:{start:%s, end:%s, prev_interval:%s, max_interval:%s, count:%s, writer:%s}" % \
               (self.start, self.end, self.prev_interval, self.max_interval, self.count, self.__writer)



#!/usr/bin/env python3

"""
Created on 19 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_timezone utility is used to shift the rec datetime field of each input document from its current offset to
the offset determined by the given timezone. This is particularly useful when using data visualisation systems that are
not localisation-aware.

When the transform is applied, the the date / time is shifted, and the UTC offset field is altered accordingly. Thus,
the resulting localised datetime represents the same point in global history, presented for a alternative UTC offset.

In most cases, the UTC offset of a timezone is dependent on daylight saving time. The sample_timezone utility is
sensitive to this - the adjustment to the UTC offset of each document's rec field is dependent on its specific
date / time value. Thus, UTC offsets are not the same thing as timezones! It is always safe to shift from UTC to a
timezone, but sifting to a subsequent timezone can have unexpected results, particularly on daylight saving time
boundaries.

The sample_timezone utility silently ignores input lines which have:
* A malformed JSON document
* No 'rec' field, or a 'rec' field with a malformed ISO 8601 datetime

Note that the timezone of a South Coast Science device is normally reported on its status topic.

SYNOPSIS
sample_timezone.py { -z | [-i ISO_PATH] TIMEZONE_NAME }

EXAMPLES
aws_topic_history.py south-coast-science-dev/production-test/loc/1/climate -s 2018-10-28T00:00:00+00:00 \
-e 2018-10-28T03:00:00+00:00 | sample_timezone.py -n Europe/Paris

DOCUMENT EXAMPLE - INPUT
{"val": {"hmd": 49.7, "tmp": 17.5}, "rec": "2018-10-28T00:00:46.037+00:00", "tag": "scs-be2-2"}

DOCUMENT EXAMPLE - OUTPUT
{"val": {"hmd": 49.7, "tmp": 17.5}, "rec": "2018-10-28T01:00:46.037+01:00", "tag": "scs-be2-2"}

RESOURCES
https://en.wikipedia.org/wiki/ISO_8601
http://pytz.sourceforge.net

BUGS
Because the sample_timezone utility is based on the pytz library, offset errors may occur when using the Etc/GMT+/-N
pseudo-timezones.
"""

import json
import sys

from scs_analysis.cmd.cmd_sample_timezone import CmdSampleTimezone

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify

from scs_core.location.timezone import Timezone


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    timezone = None
    zone = None

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleTimezone()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_timezone: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if cmd.timezone is not None:
            try:
                timezone = Timezone(cmd.timezone)
                zone = timezone.zone

            except ValueError:
                print("sample_timezone: unrecognised name:%s" % cmd.timezone, file=sys.stderr)
                exit(2)

            if cmd.verbose:
                print("sample_timezone: %s" % timezone, file=sys.stderr)
                sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.zones:
            for zone in Timezone.zones():
                print(zone, file=sys.stderr)
            exit(0)

        for line in sys.stdin:
            try:
                jdict = json.loads(line)
            except ValueError:
                continue

            document_count += 1

            try:
                rec = jdict[cmd.iso]
            except KeyError:
                continue

            # zone shift...
            datetime = LocalizedDatetime.construct_from_iso8601(rec)

            if datetime is None:
                continue

            jdict['rec'] = datetime.localize(zone).as_iso8601()

            # report...
            print(JSONify.dumps(jdict))
            sys.stdout.flush()

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_timezone: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 18 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The organisation_users utility is used to

SYNOPSIS
organisation_users.py  [-c CREDENTIALS] { -F { -e EMAIL | -l ORG_LABEL } | \
-R -e EMAIL -l ORG_LABEL | \
-C -e EMAIL -l ORG_LABEL -o { 1 | 0 } -d { 1 | 0 } | \
-U -e EMAIL -l ORG_LABEL [-o { 1 | 0 }] [-d { 1 | 0 }] [-s { 1 | 0 }] | \
-D -e EMAIL -l ORG_LABEL } \
[-i INDENT] [-v]

EXAMPLES
organisation_users.py -F -l NARA

DOCUMENT EXAMPLE
{"Username": 111, "OrgID": 1, "IsOrgAdmin": true, "IsDeviceAdmin": true, "IsSuspended": false}

SEE ALSO
scs_analysis/cognito_credentials
"""

import requests
import sys

from scs_analysis.cmd.cmd_organisation_users import CmdOrganisationUsers

from scs_core.aws.security.cognito_user_finder import CognitoUserFinder
from scs_core.aws.security.cognito_login_manager import CognitoUserLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials

from scs_core.aws.security.organisation import OrganisationUser
from scs_core.aws.security.organisation_manager import OrganisationManager

from scs_core.data.datum import Datum
from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    credentials = None
    auth = None
    cognito = None
    org = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdOrganisationUsers()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('organisation_users', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)

        if cmd.email is not None and not Datum.is_email_address(cmd.email):
            logger.error("email address '%s' is not valid." % cmd.email)
            exit(2)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        gatekeeper = CognitoUserLoginManager(requests)

        # CognitoUserCredentials...
        if not CognitoUserCredentials.exists(Host, name=cmd.credentials_name):
            logger.error("Cognito credentials not available.")
            exit(1)

        try:
            password = CognitoUserCredentials.password_from_user()
            credentials = CognitoUserCredentials.load(Host, name=cmd.credentials_name, encryption_key=password)
        except (KeyError, ValueError):
            logger.error("incorrect password")
            exit(1)

        try:
            auth = gatekeeper.login(credentials)

        except HTTPException as ex:
            logger.error(ex.data)
            exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        finder = CognitoUserFinder(requests)
        manager = OrganisationManager(requests)


        # ------------------------------------------------------------------------------------------------------------
        # validate...

        if cmd.email:
            cognito = finder.get_by_email(auth.id_token, cmd.email)

            if cognito is None:
                logger.error("no Cognito user found for email: '%s'." % cmd.email)
                exit(1)

        if cmd.org_label:
            org = manager.get_organisation_by_label(auth.id_token, cmd.org_label)

            if org is None:
                logger.error("no organisation found for label: '%s'." % cmd.org_label)
                exit(1)

        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            if cmd.email:
                report = manager.find_users_by_username(auth.id_token, cognito.username)
            else:
                report = manager.find_users_by_organisation(auth.id_token, org.org_id)

        if cmd.retrieve:
            report = manager.get_user(auth.id_token, cognito.username, org.org_id)

        if cmd.create:
            is_org_admin = cmd.org_admin == 1
            is_device_admin = cmd.device_admin == 1
            is_suspended = False

            report = OrganisationUser(cognito.username, org.org_id, is_org_admin, is_device_admin, is_suspended)
            manager.assert_user(auth.id_token, report)

        if cmd.update:
            user = manager.get_user(auth.id_token, cognito.username, org.org_id)

            if user is None:
                logger.error("no organisation user found for email: '%s' and label '%s'." % (cmd.email, cmd.org_label))
                exit(1)

            is_org_admin = user.is_org_admin if cmd.org_admin is None else bool(cmd.org_admin)
            is_device_admin = user.is_device_admin if cmd.device_admin is None else bool(cmd.device_admin)
            is_suspended = user.is_suspended if cmd.suspended is None else bool(cmd.suspended)

            report = OrganisationUser(cognito.username, org.org_id, is_org_admin, is_device_admin, is_suspended)
            manager.assert_user(auth.id_token, report)

        if cmd.delete:
            manager.delete_user(auth.id_token, cognito.username, org.org_id)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        logger.error(ex.data)
        exit(1)



"""
Created on 13 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

uses matplotlibrc configuration file

https://matplotlib.org/faq/usage_faq.html
https://stackoverflow.com/questions/28269157/plotting-in-a-non-blocking-way-with-matplotlib?noredirect=1&lq=1
"""

from matplotlib import pyplot as plt

from scs_analysis.chart.chart import Chart


# --------------------------------------------------------------------------------------------------------------------

class MultiChart(Chart):
    """
    classdocs
    """

    def __init__(self, title, batch_mode, x_count, y_min, y_max, *paths):
        """
        Constructor
        """
        Chart.__init__(self, title, batch_mode)

        # fields...
        self.__y_min = y_min
        self.__y_max = y_max

        self.__paths = paths

        # data...
        self.__x_data = range(x_count)

        self.__y_data = self.__init_data(x_count)

        # plotter...
        fig = plt.figure()
        fig.canvas.set_window_title(self.title(', '.join(self.__paths)))
        fig.tight_layout()

        fig.canvas.mpl_connect('close_event', self.close)

        ax1 = plt.axes()
        ax1.xaxis.grid(True)

        plt.grid()

        plt.ylim([y_min, y_max])

        if not batch_mode:
            plot_data = self.__plot_data(self.__x_data)
            self.__lines = plt.plot(*plot_data)


    # ----------------------------------------------------------------------------------------------------------------

    def plot(self, path_dict):
        if self.closed:
            return

        # datum...
        datum = []

        for path in self.__paths:
            try:
                value = path_dict.node(path)
            except KeyError:
                return

            if value is None:
                return

            datum.append(float(value))

        # update data...
        for i in range(len(datum)):
            del self.__y_data[i][0]
            self.__y_data[i].append(datum[i])

            if not self.batch_mode:
                self.__lines[i].set_ydata(self.__y_data[i])

        # min / max...
        min_data_y = float(min(self.__flat_y_data))
        max_data_y = float(max(self.__flat_y_data))

        min_axis_y = self.__y_min if min_data_y > self.__y_min else min_data_y - (abs(min_data_y) * 0.05)
        max_axis_y = self.__y_max if max_data_y < self.__y_max else max_data_y + (abs(max_data_y) * 0.05)

        plt.ylim([min_axis_y, max_axis_y])

        if not self.batch_mode:
            self.render(block=False)


    def render(self, block=True):
        if self.batch_mode:
            plot_data = self.__plot_data(self.__x_data)
            plt.plot(*plot_data)

        plt.draw()
        plt.show(block=block)
        plt.pause(0.001)


    def close(self, _event):
        self._closed = True
        plt.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __init_data(self, x_count):
        data = []

        for _ in range(len(self.__paths)):
            data.append([self.__y_min] * x_count)

        return data


    def __plot_data(self, x_data):
        coordinates = []

        for i in range(len(self.__y_data)):
            coordinates.append(x_data)
            coordinates.append(self.__y_data[i])

        return coordinates


    @property
    def __flat_y_data(self):
        return [float(item) for sublist in self.__y_data for item in sublist]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def y_min(self):
        return self.__y_min


    @property
    def y_max(self):
        return self.__y_max


    @property
    def y_data(self):
        return self.__y_data


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MultiChart:{batch_mode:%s, y_min:%s, y_max:%s, paths:%s}" % \
                (self.batch_mode, self.y_min, self.y_max, self.__paths)



"""
Created on 6 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdAWSTopicPublisher(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog -t TOPIC [-v]", version="%prog 1.0")

        # compulsory...
        self.__parser.add_option("--topic", "-t", type="string", nargs=1, action="store", dest="topic",
                                 help="topic path")

        # optional...
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__opts.topic


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdAWSTopicPublisher:{topic:%s, verbose:%s}" %  (self.topic, self.verbose)



"""
Created on 4 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdHistoChart(object):
    """
    unix command line handler
    """

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-b] [-x MIN MAX] [-c BIN_COUNT] [-p PRECISION] "
                                                    "[-o FILENAME] [-e] [-t TITLE] [-v] PATH", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--batch", "-b", action="store_true", dest="batch_mode", default=False,
                                 help="wait for all data before displaying chart")

        self.__parser.add_option("--x", "-x", type="float", nargs=2, action="store", default=(-1.0, 1.0), dest="x",
                                 help="set x-axis to min / max (default -1, 1)")

        self.__parser.add_option("--bincount", "-c", type="int", nargs=1, action="store", default=200, dest="bin_count",
                                 help="number of bins (default 200)")

        self.__parser.add_option("--precision", "-p", type="int", nargs=1, action="store", default=3, dest="precision",
                                 help="precision of reported deltas (default 3)")

        self.__parser.add_option("--skip-malformed", "-s", action="store_true", dest="skip_malformed", default=False,
                                 help="ignore rows with missing path values")

        self.__parser.add_option("--output", "-o", type="string", nargs=1, action="store", dest="outfile",
                                 help="output histogram to CSV file")

        self.__parser.add_option("--echo", "-e", action="store_true", dest="echo", default=False,
                                 help="echo stdin to stdout")

        self.__parser.add_option("--title", "-t", type="string", nargs=1, action="store", dest="title",
                                 help="chart title")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.path is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def batch_mode(self):
        return self.__opts.batch_mode


    @property
    def x(self):
        return self.__opts.x


    @property
    def bin_count(self):
        return self.__opts.bin_count


    @property
    def precision(self):
        return self.__opts.precision


    @property
    def outfile(self):
        return self.__opts.outfile


    @property
    def echo(self):
        return self.__opts.echo


    @property
    def skip_malformed(self):
        return self.__opts.skip_malformed


    @property
    def title(self):
        return self.__opts.title


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdHistoChart:{batch_mode:%s, x:%s, bin_count:%d, precision:%d, outfile:%s, echo:%s, " \
               "skip_malformed:%s, title:%s, verbose:%s, path:%s}" % \
                    (self.batch_mode, self.x, self.bin_count, self.precision, self.outfile, self.echo,
                     self.skip_malformed, self.title, self.verbose, self.path)



#!/usr/bin/env python3

"""
Created on 20 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The csv_collator utility is used to separate the input JSON documents according to the upper and lower bounds of a
sequence of bins. For each bin, assignment follows the rule:

lower bound <= value < upper bound

The upper and lower bounds for the data set should be specified, along with a step size. The number of bins required
to service this domain is calculated automatically. The path identifying the leaf node in the input document where the
value is to be found must be specified.

File names (and sub-directories) as specified by the --file-prefix flag.

Documents that do not contain a field at the specified path, or have values that cannot be evaluated as a float, are
ignored. Likewise, values outside the upper and lower bounds are ignored.

On completion, a summary of the bin assignments is written to stdout.  If no file prefix is given, then the CSV files
are not generated, but the report is still produced.

Two collators are provided in this package: csv_collator collates into separate CSV files (collate to rows),
whereas sample_collator collates into separate columns (collate to columns).

SYNOPSIS
csv_collator.py -l LOWER_BOUND -u UPPER_BOUND -d DELTA [-f FILE_PREFIX] [-v] PATH

EXAMPLES
csv_reader.py alphasense_303_2018-08.csv | \
csv_collator.py -l 5.0 -u 21.0 -d 1.0 -f collation/alphasense_303_2018-08 -v val.sht.hmd.aH

FILES
Output file names are of the form:
FILE-PREFIX_DOMAIN-LOW_DOMAIN-HIGH.csv

SEE ALSO
scs_analysis/csv_collation_summary
scs_analysis/csv_segmentor
scs_analysis/sample_collator

"""

import sys

from scs_analysis.cmd.cmd_csv_collator import CmdCSVCollator
from scs_analysis.handler.csv_collator import CSVCollator

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    collator = None

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdCSVCollator()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("csv_collator: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        collator = CSVCollator.construct(cmd.lower, cmd.upper, cmd.delta, cmd.file_prefix)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # receive...
        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                break

            document_count += 1

            if cmd.path not in datum.paths():
                continue

            try:
                value = float(datum.node(cmd.path))
            except (TypeError, ValueError):
                continue

            if not collator.collate(value, line):
                continue

            processed_count += 1

        # report...
        for b in collator.bins:
            print(JSONify.dumps(b))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("csv_collator: KeyError: %s" % ex, file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if collator is not None:
            collator.close()

        if cmd.verbose:
            print("csv_collator: documents: %d processed: %d" % (document_count, processed_count),
                  file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 6 Nov 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_error utility is used to compute either the difference between or the ratio of a reported value and a
reference value for a given path is a stream of JSON documents.

Input is a sequence of JSON documents on stdin, containing both reported and reference values. If the specified
REFERENCE_PATH and REPORTED_PATH are not present, the utility terminates. If either value cannot be interpreted as a
floating-point value, the document is ignored. If the specified ERROR_PATH is already present in the document, it is
overwritten.

SYNOPSIS
sample_error.py { -l | -s } [-p PRECISION] [-v] REFERENCE_PATH REPORTED_PATH ERROR_PATH

EXAMPLES
csv_reader.py -v Pi-R1-joined-2019-10-15min.csv | \
sample_error.py -s -v "fidas.PM25 Converted Measurement" r1.val.pm2p5 error.pm2p5 | \
sample_error.py -s -v "fidas.PM10 Converted Measurement" r1.val.pm10 error.pm10 | \
csv_writer.py -v Pi-R1-joined-2019-10-15min-error.csv

DOCUMENT EXAMPLE - INPUT
{"rec": "2019-09-24T13:15:00Z", "fidas": {"Site": "Heathrow LHR2", "SiteCode": "LHR2",
"PM10 Processed Measurement": 6.3, "PM10 Converted Measurement": 6.3, "PM10 Data Status": "Provisional",
"PM25 Processed Measurement": 3.447, "PM25 Converted Measurement": 3.2519, "PM25 Data Status": "Provisional"},
"r1": {"val": {"mtf1": 32.4, "pm1": 2.1, "mtf5": 34.1, "pm2p5": 6.3, "sht": {"hmd": 29.5, "tmp": 33.0}},
"tag": "scs-pb1-3", "src": "R1"}}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2019-09-24T13:15:00Z", "fidas": {"Site": "Heathrow LHR2", "SiteCode": "LHR2",
"PM10 Processed Measurement": 6.3, "PM10 Converted Measurement": 6.3, "PM10 Data Status": "Provisional",
"PM25 Processed Measurement": 3.447, "PM25 Converted Measurement": 3.2519, "PM25 Data Status": "Provisional"},
"r1": {"val": {"mtf1": 32.4, "pm1": 2.1, "mtf5": 34.1, "pm2p5": 6.3, "sht": {"hmd": 29.5, "tmp": 33.0}},
"tag": "scs-pb1-3", "src": "R1"},
"error": {"pm2p5": 1.937, "pm10": 1.889}}
"""

import sys

from scs_analysis.cmd.cmd_sample_error import CmdSampleError

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleError()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_error: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        max_datum = None

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            document_count += 1

            # reference...
            if cmd.reference_path not in datum.paths():
                print("sample_error: reference path '%s' not present" % cmd.reference_path, file=sys.stderr)
                exit(1)

            try:
                reference = float(datum.node(cmd.reference_path))
            except (TypeError, ValueError):
                continue

            # reported...
            if cmd.reported_path not in datum.paths():
                print("sample_error: reported path '%s' not present" % cmd.reference_path, file=sys.stderr)
                exit(1)

            try:
                reported = float(datum.node(cmd.reported_path))
            except (TypeError, ValueError):
                continue

            # error...
            try:
                error = reported / reference if cmd.scaling else reported - reference
            except ZeroDivisionError:
                continue

            # report...
            datum.append(cmd.error_path, round(error, cmd.precision))
            sys.stdout.flush()

            processed_count += 1

            print(JSONify.dumps(datum))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_error: KeyError: %s" % ex, file=sys.stderr)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_error: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 11 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The node utility is used to extract a node or nodes from within a JSON document. Data is presented as a sequence of
documents on stdin, and the extracted node(s) are passed to stdout. Alternatively to stdin, a single JSON document
can be read from a file.The extracted node may be a leaf node or an internal node.

By default, only the specified nodes are passed to the output. In the --exclude mode, all nodes are passed to stdout,
with the exception of the specified nodes. In the default mode, if no node path is specified, the whole input document
is passed to stdout. In the --exclude mode, if no node path is specified, then nothing is output.

By default, output is in the form of a sequence of JSON documents, separated by newlines. If the array (-a) option is
selected, output is in the form of a JSON array - the output opens with a '[' character, documents are separated by
the ',' character, and the output is terminated by a ']' character.

Alternatively, if the node is an array or other iterable type, then it may be output as a sequence (a list of items
separated by newline characters) according to the -s flag.

The ordering of output nodes is as follows:
* Exclude mode - by input nodes
* Default mode - by nodes specified on the command line

WARNING: node ordering is overridden by the internal node structure. Thus the specified ordering: a.b.c, x.b.c, a.b.d
would be rendered as: a.b.c, a.b.d, x.b.c

SYNOPSIS
node.py [{ [-x] [-a] | -s }] [-f FILE] [-i INDENT] [-v] [SUB_PATH_1 .. SUB_PATH_N]

EXAMPLES
csv_reader.py climate.csv | node.py -x val.bar

DOCUMENT EXAMPLE - INPUT
{"val": {"hmd": 73.5, "tmp": 10.8, "bar": ""}, "rec": "2019-02-17T08:56:53Z"}
{"val": {"hmd": 73.6, "tmp": 10.8, "bar": ""}, "rec": "2019-02-17T08:57:53Z"}

DOCUMENT EXAMPLE - OUTPUT
default mode:
{"val": {"hmd": 73.5, "tmp": 10.8}, "rec": "2019-02-17T08:56:53Z"}
{"val": {"hmd": 73.6, "tmp": 10.8}, "rec": "2019-02-17T08:57:53Z"}

array mode:
[{"val": {"hmd": 73.5, "tmp": 10.8}, "rec": "2019-02-17T08:56:53Z"},
{"val": {"hmd": 73.6, "tmp": 10.8}, "rec": "2019-02-17T08:57:53Z"}]
"""

import json
import sys

from scs_analysis.cmd.cmd_node import CmdNode

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    output_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdNode()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('node', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if cmd.filename:
            try:
                with open(cmd.filename, 'r') as file:
                    source = [file.read()]

            except FileNotFoundError:
                logger.error("file not found: %s" % cmd.filename)
                exit(1)

        else:
            source = sys.stdin

        if cmd.sequence:
            for document in source:
                try:
                    source = [json.dumps(item) for item in json.loads(document)]
                except json.JSONDecodeError:
                    logger.error("invalid document: '%s'" % document.strip())
                    exit(1)
                break


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.array:
            print('[', end='')

        node = None
        first = True

        for document in source:
            datum = PathDict.construct_from_jstr(document)

            if datum is None:
                continue

            document_count += 1

            if cmd.exclude and not cmd.has_sub_paths():
                continue                                # everything is excluded

            # build...
            if not cmd.has_sub_paths():
                target = datum                          # everything is included

            else:
                target = PathDict()

                if cmd.exclude:
                    # use datum field ordering...
                    for path in datum.paths():
                        if cmd.includes(path):
                            target.append(path, datum.node(path))

                else:
                    # use cmd.sub_paths field ordering...
                    for sub_path in cmd.sub_paths:
                        if datum.has_sub_path(sub_path):
                            target.append(sub_path, datum.node(sub_path))

            # report...
            if not target:
                continue                                # skip empty outputs

            if cmd.array:
                if first:
                    print(JSONify.dumps(target), end='')
                    first = False

                else:
                    print(", %s" % JSONify.dumps(target), end='')

            else:
                print(JSONify.dumps(target, indent=cmd.indent))
                sys.stdout.flush()
                output_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        logger.error("KeyError: %s" % ex)
        exit(1)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.array:
            print(']')
            output_count = 1

        logger.info("documents: %d output: %d" % (document_count, output_count))
